//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2007-10-11
//
#ifndef IEEE_80211_MAC_LBMS_H
#define IEEE_80211_MAC_LBMS_H

// uncomment this if you do not want to log state machine transitions
#define FSM_DEBUG

#include <list>
#include <vector>
#include "Ieee80211Mac.h"
#include "WirelessMacBase.h"
#include "IPassiveQueue.h"
#include "Ieee80211Frame_m.h"
#include "Ieee80211Consts.h"
#include "NotificationBoard.h"
#include "RadioState.h"
#include "FSMA.h"

class INET_API LbmsMacAP : public Ieee80211Mac
{

  enum 
  {
      SNR,					// The leader is the station with the worst SNR
      PACKET_LOST_RATE		// The Leader is the station 
  }leaderElection;
   enum 
  {
      Worst,	// The leader is the station who has the worst recption condition
      Medium,	// The leader is the station who has the medium recption condition
	  Best		// The leader is the station who has the best recption condition 
  }leaderChoice;

  struct MulticastMACAddress
  {
      MACAddress address;		// MAC address of the station
	  bool isLeader;			// true is the station is the Leader
      double packetloss;		// The packet loss of the station
      double snr;				// ths SNR of the station
  };

  typedef std::vector<MulticastMACAddress> MulticastMACAddressList;
  typedef std::vector<MulticastMACAddress>::iterator MulticastMACAddressListI;

  struct MulticastGroup
  {
     MACAddress groupAddress;
     MulticastMACAddressList MAClist;
  };

  typedef std::vector<MulticastGroup> MulticastGroupList;
  typedef std::vector<MulticastGroup>::iterator MulticastGroupListI;

  protected:
    /**
     * @name Ieee80211Mac state variables
     * Various state information checked and modified according to the state machine.
     */
    //@{
    /**
     * A list MAC group address whitch MAC is associated with 
     */
    MulticastGroupList MulticastGrouplist;
    //@}

  protected:
    /** @name Statistics */
    //@{
    long numReceivedMulticast;
    //@}

  protected:
    /**
     * @name Initialization functions
     */
    //@{
    /** @brief Initialization of the module and its variables */
    virtual int numInitStages() const {return 2;}
    virtual void initialize(int);
    //@}

    protected:
    /**
     * @name Message handing functions
     * @brief Functions called from other classes to notify about state changes and to handle messages.
     */
    //@{
    /** @brief Handle messages from upper layer */
    virtual void handleUpperMsg(cMessage *msg);
    /** @brief Handle all kinds of messages and notifications with the state machine */
    virtual void handleWithFSM(cMessage *msg);
    //@}

    protected:
    /**
     * @name Timer functions
     * @brief These functions have the side effect of starting the corresponding timers.
     */
    //@{
    void scheduleMulticastTimeoutPeriod(Ieee80211DataOrMgmtFrame *frame);
    //@}

  protected:
    /**
     * @name Frame transmission functions
     */
    //@{
	void sendMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend);
    void sendLBMSReportFrame(Ieee80211DataOrMgmtFrame*);
    void sendLbmsReport(Ieee80211Frame *mframe);
    void sendLbmsReport(bool leader, MACAddress &groupAddress, MACAddress &memberAddress);
    void sendLbmsReport();
    //@}
      protected:
    /**
     * @name Frame builder functions
     */
    //@{
    Ieee80211DataOrMgmtFrame *buildMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend);
    //@}

  protected:
    /**
     * @name Utility functions
     */
    //@{
    /** @brief returns true if message is a multicast message */
    bool isMulticast(Ieee80211Frame *msg); 

    /** @brief returns true if message destination address is ours */
    bool isForUs(Ieee80211Frame *msg);

    /** @brief returns true if the MAC multicast address is a new one */
    bool isNewGroup(Ieee80211Frame *frameToSend);

	/** @brief returns true if the MAC multicast is new member for a group multipoint */
    bool isNewMember(Ieee80211Frame *frame);

	/** @brief returns true if the MAC multicast address is a new one */
    bool isRegistredGroup(Ieee80211DataOrMgmtFrame *frame);

    /** @brief returns true if there is one or many receiver for this multicast frame */
    bool isAssociatedGroup(Ieee80211DataOrMgmtFrame *frameToSend);

    /** @brief select a new Leader */
    void setLeader(Ieee80211DataOrMgmtFrame *frameToSend);

    /** @brief returns Select a new Leader */
    void setNonLeader(Ieee80211DataOrMgmtFrame *frameToSend);

    /** @brief process Incoming LBMS frame */
    void processIncomingLBMS(Ieee80211Frame *frameToSend);

    /** @brief process Incoming LBMS report frame */
    void processLBMSReportFrame(Ieee80211Frame *frameToSend);

    /** @brief process Incoming LBMS report frame */
    void processLBMSRequestFrame(Ieee80211Frame *frameToSend);

    /** @brief process Incoming Multicast frame */
    void processIncomingMulticastFrame(Ieee80211Frame *frameToSend);

    /** @brief add a new multicast group */
    void addNewMulticastGroup(Ieee80211Frame *frameToSend);

    /** @brief add a new multicast group member */
    void addNewMulticastMember(Ieee80211Frame *frameToSend);

    /** @brief remove a new multicast group member */
    void RemoveMulticastMember(Ieee80211DataOrMgmtFrame *frame);

    /** @brief elect a new Leader for multicast group*/
    void electNewLeader(Ieee80211LBMSRequest *frame);

    /** @brief choose the best Leader for multicast group*/
    int chooseLeader(Ieee80211LBMSRequest *mframe);

    /** @brief logs all Multicast Group */
    void myMulticastGroup();
    //@}
};

#endif
