//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2007-10-11
//
#include "LbmsMacSTA.h"
#include "RadioState.h"
#include "InterfaceTable.h"
#include "InterfaceTableAccess.h"
#include "PhyControlInfo_m.h"

Define_Module(LbmsMacSTA);


LbmsMacSTA::~LbmsMacSTA()
{
    delete Leader;
}

/****************************************************************
 * Initialization functions.
 */
void LbmsMacSTA::initialize(int stage)
{
    Ieee80211Mac::initialize(stage);

    if (stage == 0)
    {
        EV << "Initializing stage 0\n";
        // state variables
        fsm.setName("LbmsMacSTA State Machine");
        leaderID = par("ID");
        Leader = new cOutVector("Leader");
        Leader->record(0);
        // statistics
        numReceivedMulticast = 0;
        WATCH(numReceivedMulticast);
    }
}

/****************************************************************
 * Message handling functions.
 */

void LbmsMacSTA::handleUpperMsg(cMessage *msg)
{
    EV<<"Ieee80211Mac handle message from upper layer"<<endl;
    // check if it's a command from the mgmt layer
    if (msg->length()==0 && msg->kind()!=0)
    {
        handleCommand(msg);
        return;
    }

    // check for queue overflow
    if (maxQueueSize && transmissionQueue.size() == maxQueueSize)
    {
        EV << "message " << msg << " received from higher layer but MAC queue is full, dropping message\n";
        delete msg;
        return;
    }


    // must be a Ieee80211DataOrMgmtFrame, within the max size because we don't support fragmentation
    Ieee80211DataOrMgmtFrame *frame = check_and_cast<Ieee80211DataOrMgmtFrame *>(msg);
    if (frame->byteLength() > fragmentationThreshold)
        error("message from higher layer (%s)%s is too long for 802.11b, %d bytes (fragmentation is not supported yet)",
              msg->className(), msg->name(), msg->byteLength());
    EV << "frame " << frame << " received from higher layer, receiver = " << frame->getReceiverAddress() << endl;

    ASSERT(!frame->getReceiverAddress().isUnspecified());

    if (!strcmp(frame->name(), "Join"))
    {
        joinMulticastGroup(frame);
        frame->setType(ST_LBMS_REQUEST);
    }
    else if (!strcmp(frame->name(), "Leave"))
    {
        leaveMulticastGroup(frame);
        frame->setType(ST_LBMS_REQUEST);
    }
    else if (!strcmp(frame->name(), "LbmsReport"))
    {
        EV<<"LbmsReport Frame"<<endl;
        Ieee80211LBMSReport *lbmsReport = check_and_cast<Ieee80211LBMSReport *>(frame);
        lbmsReport->getElement(0).multicastGroupAddress = multicastGroupList.begin()->address;
        lbmsReport->setSnr(snr);
        lbmsReport->setPacketLoss(lossRate);
        lbmsReport->setTransmitterAddress(address);
        lbmsReport->setSequenceNumber(sequenceNumber);
        sequenceNumber = (sequenceNumber+1) % 4096;  //XXX seqNum must be checked upon reception of frames!

        transmissionQueue.push_back(lbmsReport);
        handleWithFSM(lbmsReport);
        return;
    }

    // fill in missing fields (receiver address, seq number), and insert into the queue
    frame->setTransmitterAddress(address);
    frame->setSequenceNumber(sequenceNumber);
    sequenceNumber = (sequenceNumber+1) % 4096;  //XXX seqNum must be checked upon reception of frames!

    transmissionQueue.push_back(frame);

    handleWithFSM(frame);
}

/**
 * Msg can be upper, lower, self or NULL (when radio state changes)
 */
void LbmsMacSTA::handleWithFSM(cMessage *msg)
{
    // skip those cases where there's nothing to do, so the switch looks simpler
    if (isUpperMsg(msg) && fsm.state() != IDLE)
    {
        EV << "deferring upper message transmission in " << fsm.stateName() << " state\n";
        return;
    }

    Ieee80211Frame *frame = dynamic_cast<Ieee80211Frame*>(msg);

	if (msg && isLowerMsg(msg))
	{
		AirFrame *airFrame = (AirFrame *) (msg);
		frame->setPReceived(rPower);
	}

    int frameType = frame ? frame->getType() : -1;
    int msgKind = msg->kind();
    logState();
    stateVector.record(fsm.state());

    if (frame && isLowerMsg(frame))
    {
        lastReceiveFailed =(msgKind == COLLISION || msgKind == BITERROR);
        scheduleReservePeriod(frame);
    }

    FSMA_Switch(fsm)
    {
        FSMA_State(IDLE)
        {
            FSMA_Enter(sendDownPendingRadioConfigMsg());
            FSMA_Event_Transition(Data-Ready,
                                  isUpperMsg(msg),
                                  DEFER,
                invalidateBackoffPeriod();
            );
            FSMA_No_Event_Transition(Immediate-Data-Ready,
                                     !transmissionQueue.empty(),
                                     DEFER,
                invalidateBackoffPeriod();
            );
            FSMA_Event_Transition(Receive,
                                  isLowerMsg(msg),
                                  RECEIVE,
            );
        }
        FSMA_State(DEFER)
        {
            FSMA_Enter(sendDownPendingRadioConfigMsg());
            FSMA_Event_Transition(Wait-DIFS,
                                  isMediumStateChange(msg) && isMediumFree(),
                                  WAITDIFS,
            ;);
            FSMA_No_Event_Transition(Immediate-Wait-DIFS,
                                     isMediumFree() || !backoff,
                                     WAITDIFS,
            ;);
            FSMA_Event_Transition(Receive,
                                  isLowerMsg(msg),
                                  RECEIVE,
            ;);
        }
        FSMA_State(WAITDIFS)
        {
            FSMA_Enter(scheduleDIFSPeriod());
            FSMA_Event_Transition(Immediate-Transmit-RTS,
                                  msg == endDIFS && !isBroadcast(currentTransmission()) && !isMulticast(currentTransmission())
                                  && currentTransmission()->byteLength() >= rtsThreshold && !backoff,
                                  WAITCTS,
                sendRTSFrame(currentTransmission());
                cancelDIFSPeriod();
            );//ahmed

            FSMA_Event_Transition(Immediate-Transmit-Broadcast,
                                  msg == endDIFS && isBroadcast(currentTransmission()) && !backoff,
                                  WAITBROADCAST,
                sendBroadcastFrame(currentTransmission());
                cancelDIFSPeriod();
            );
            FSMA_Event_Transition(Immediate-Transmit-Multicast,
                                  msg == endDIFS && isMulticast(currentTransmission()) && !backoff,
                                  WAITMULTICAST,
                sendMulticastFrame(currentTransmission());
                cancelDIFSPeriod();
            );//sendind mutlicast form upper layer
            FSMA_Event_Transition(Immediate-Transmit-Data,
                                  msg == endDIFS && !isBroadcast(currentTransmission()) && !isMulticast(currentTransmission()) && !backoff,
                                  WAITACK,
                sendDataFrame(currentTransmission());//ahmed
                cancelDIFSPeriod();
            );
            FSMA_Event_Transition(DIFS-Over,
                                  msg == endDIFS,
                                  BACKOFF,
                ASSERT(backoff);
                if (isInvalidBackoffPeriod())
                    generateBackoffPeriod();
            );
            FSMA_Event_Transition(Busy,
                                  isMediumStateChange(msg) && !isMediumFree(),
                                  DEFER,
                backoff = true;
                cancelDIFSPeriod();
            );
            FSMA_No_Event_Transition(Immediate-Busy,
                                     !isMediumFree(),
                                     DEFER,
                backoff = true;
                cancelDIFSPeriod();
            );
            // radio state changes before we actually get the message, so this must be here
            FSMA_Event_Transition(Receive,
                                  isLowerMsg(msg),
                                  RECEIVE,
                cancelDIFSPeriod();
            ;);
        }
        FSMA_State(BACKOFF)
        {
            FSMA_Enter(scheduleBackoffPeriod());
            FSMA_Event_Transition(Transmit-RTS,
                                  msg == endBackoff && !isBroadcast(currentTransmission()) && !isMulticast(currentTransmission())
                                  && currentTransmission()->byteLength() >= rtsThreshold,
                                  WAITCTS,
                sendRTSFrame(currentTransmission());
            );//ahmed
            FSMA_Event_Transition(Transmit-Broadcast,
                                  msg == endBackoff && isBroadcast(currentTransmission()),
                                  WAITBROADCAST,
                sendBroadcastFrame(currentTransmission());
            );
            FSMA_Event_Transition(Transmit-Multicast,
                                  msg == endBackoff && isMulticast(currentTransmission()),
                                  WAITMULTICAST,
                sendMulticastFrame(currentTransmission());
            );//sendind mutlicast form upper layer
            FSMA_Event_Transition(Transmit-Data,
                                  msg == endBackoff && !isBroadcast(currentTransmission()) && !isMulticast(currentTransmission()),
                                  WAITACK,
                sendDataFrame(currentTransmission());
            );//ahmed
            FSMA_Event_Transition(Backoff-Busy,
                                  isMediumStateChange(msg) && !isMediumFree(),
                                  DEFER,
                cancelBackoffPeriod();
                decreaseBackoffPeriod();
            );
        }
        FSMA_State(WAITACK)
        {
            FSMA_Enter(scheduleDataTimeoutPeriod(currentTransmission()));
            FSMA_Event_Transition(Receive-ACK,
                                  isLowerMsg(msg) && isForUs(frame) && frameType == ST_ACK,
                                  IDLE,
                cancelTimeoutPeriod();
                finishCurrentTransmission();
                if (retryCounter == 0) numSentWithoutRetry++;
                numSent++;
            );
            FSMA_Event_Transition(Transmit-Data-Failed,
                                  msg == endTimeout && retryCounter == retryLimit,
                                  IDLE,
                giveUpCurrentTransmission();
            );
            FSMA_Event_Transition(Receive-ACK-Timeout,
                                  msg == endTimeout,
                                  DEFER,
                retryCurrentTransmission();
            );
        }
        // wait until broadcast is sent
        FSMA_State(WAITBROADCAST)
        {
            FSMA_Enter(scheduleBroadcastTimeoutPeriod(currentTransmission()));
            FSMA_Event_Transition(Transmit-Broadcast,
                                  msg == endTimeout,
                                  IDLE,
                finishCurrentTransmission();
                numSentBroadcast++;
            );
        }

        FSMA_State(WAITMULTICAST)
        {
            FSMA_Enter(scheduleMulticastTimeoutPeriod(currentTransmission()));
            FSMA_Event_Transition(Transmit-Multicast,
                                  msg == endTimeout,
                                  IDLE,
                finishCurrentTransmission();
            );
        }//ahmed
        // accoriding to 9.2.5.7 CTS procedure
        FSMA_State(WAITCTS)
        {
            FSMA_Enter(scheduleCTSTimeoutPeriod());
            FSMA_Event_Transition(Receive-CTS,
                                  isLowerMsg(msg) && isForUs(frame) && frameType == ST_CTS,
                                  WAITSIFS,
                cancelTimeoutPeriod();
            );
            FSMA_Event_Transition(Transmit-RTS-Failed,
                                  msg == endTimeout && retryCounter == retryLimit,
                                  IDLE,
                giveUpCurrentTransmission();
            );
            FSMA_Event_Transition(Receive-CTS-Timeout,
                                  msg == endTimeout,
                                  DEFER,
                retryCurrentTransmission();
            );
        }
        FSMA_State(WAITSIFS)
        {
            FSMA_Enter(scheduleSIFSPeriod(frame));
            FSMA_Event_Transition(Transmit-CTS,
                                  msg == endSIFS && frameReceivedBeforeSIFS()->getType() == ST_RTS,
                                  IDLE,
                sendCTSFrameOnEndSIFS();
                resetStateVariables();
            );
            FSMA_Event_Transition(Transmit-DATA,
                                  msg == endSIFS && frameReceivedBeforeSIFS()->getType() == ST_CTS,
                                  WAITACK,
                sendDataFrameOnEndSIFS(currentTransmission());
            );
            FSMA_Event_Transition(Transmit-ACK,
                                  msg == endSIFS && isDataOrMgmtFrame(frameReceivedBeforeSIFS()),
                                  IDLE,
                sendACKFrameOnEndSIFS();
                resetStateVariables();
            );
        }
        // this is not a real state
        FSMA_State(RECEIVE)
        {
            FSMA_No_Event_Transition(Immediate-Receive-Error,
                                     isLowerMsg(msg) && (msgKind == COLLISION || msgKind == BITERROR),
                                     IDLE,
                EV << "received frame contains bit errors or collision, next wait period is EIFS\n";
                numCollision++;
                resetStateVariables();
            );
            FSMA_No_Event_Transition(Immediate-Receive-Broadcast,
                                     isLowerMsg(msg) && isBroadcast(frame) && isDataOrMgmtFrame(frame),
                                     IDLE,
                sendUp(frame);
                numReceivedBroadcast++;
                resetStateVariables();
            );
            FSMA_No_Event_Transition(Immediate-Receive-LBMS,
                                     isLowerMsg(msg) && isForUs(frame) && (frameType == ST_LBMS_REPORT || frameType == ST_LBMS_REQUEST),
                                     WAITSIFS,
                 processIncomingLBMS(frame);
            );
            FSMA_No_Event_Transition(Immediate-Receive-Multicast,
                                     isLowerMsg(msg) && isMulticast(frame) && isForUs(frame) && isLeader(frame) && isDataOrMgmtFrame(frame),
                                     WAITSIFS,
                sendUp(frame);
                numReceivedMulticast++;
                resetStateVariables();
            );
            FSMA_No_Event_Transition(Immediate-Receive-Multicast,
                                     isLowerMsg(msg) && isMulticast(frame) && isForUs(frame) && !isLeader(frame) && isDataOrMgmtFrame(frame),
                                     IDLE,
                sendUp(frame);
                numReceivedMulticast++;
                resetStateVariables();
            );
            FSMA_No_Event_Transition(Immediate-Receive-Data,
                                     isLowerMsg(msg) && isForUs(frame) && isDataOrMgmtFrame(frame),
                                     WAITSIFS,
                sendUp(frame);
                numReceived++;
            );
            FSMA_No_Event_Transition(Immediate-Receive-RTS,
                                     isLowerMsg(msg) && isForUs(frame) && frameType == ST_RTS,
                                     WAITSIFS,
				 delete frame;
            );
            FSMA_No_Event_Transition(Immediate-Receive-Other,
                                     isLowerMsg(msg),
                                     IDLE,
                resetStateVariables();
            );
        }
    }

    logState();
    stateVector.record(fsm.state());
}


   
/****************************************************************
 * Timer functions.
 */

void LbmsMacSTA::scheduleMulticastTimeoutPeriod(Ieee80211DataOrMgmtFrame *frameToSend)
{
    EV << "scheduling multicast timeout period\n";
    scheduleAt(simTime() + frameDuration(frameToSend), endTimeout);
}

/****************************************************************
 * Frame sender functions.
 */


void LbmsMacSTA::sendMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend)
{
     sendDown(buildMulticastFrame(frameToSend));
}

/****************************************************************
 * Frame builder functions.
 */
Ieee80211DataOrMgmtFrame *LbmsMacSTA::buildMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend)
{
    Ieee80211DataOrMgmtFrame *frame = (Ieee80211DataOrMgmtFrame *)frameToSend->dup();
    frame->setDuration(0);
    return frame;
}

/****************************************************************
 * Helper functions.
 */

bool LbmsMacSTA::isMulticast(Ieee80211Frame *frame)
{	
	if (frame && (frame->getReceiverAddress().isMulticast()))
		return true;
	else 
	{
		Ieee80211DataOrMgmtFrame *mframe = (Ieee80211DataOrMgmtFrame *) frame;
		if (mframe && (mframe->getAddress3().isMulticast()))
			return true;
		else
			return false;
	}
}

bool LbmsMacSTA::isForUs(Ieee80211Frame *frame)
{
    if (frame && frame->getReceiverAddress() == address) 
	return true;
	else 
	{
		Ieee80211DataOrMgmtFrame *mframe = (Ieee80211DataOrMgmtFrame *) frame;
		for (MulticastMACAddressListI j = multicastGroupList.begin (); j != multicastGroupList.end (); j++)
		{
			if (j->address == frame->getReceiverAddress())
			{
				return true;
			}
		}
	}
	return false;
}

bool LbmsMacSTA::isLeader(Ieee80211Frame *frame)
{
    for (MulticastMACAddressListI j = multicastGroupList.begin (); j != multicastGroupList.end (); j++)
    {
         if (j->address == frame->getReceiverAddress()) 
         {
             if (j->isLeader)
             Leader->record(leaderID);
             return j->isLeader;
         }
    }
    return false;
}
void LbmsMacSTA::setLeader(Ieee80211Frame *frame)
{
    EV <<" ######################################################################## "<<endl;
    Ieee80211LBMSReport *mframe = check_and_cast<Ieee80211LBMSReport *>(frame);
    EV <<parentModule()->fullPath()<<"."<<className()<<" Groupe Address " <<mframe->getElement(0).multicastGroupAddress<<endl;
    EV <<parentModule()->fullPath()<<"."<<className()<<" SetLeader "<<mframe->getElement(0).leader <<endl;
    for (MulticastMACAddressListI j = multicastGroupList.begin (); j != multicastGroupList.end (); j++) {
         if (j->address == mframe->getElement(0).multicastGroupAddress) j->isLeader = mframe->getElement(0).leader;
    }
    EV <<" ######################################################################## "<<endl;
}

void LbmsMacSTA::joinMulticastGroup(Ieee80211DataOrMgmtFrame *frame)
{
    Ieee80211LBMSRequest *mframe = (Ieee80211LBMSRequest *) frame;

    MulticastMACAddress multicastGroup;
    multicastGroup.address =  mframe->getElement().multicastAddress;
    multicastGroup.isLeader = false;
    multicastGroupList.push_back(multicastGroup);
    logMultiCast();
}

void LbmsMacSTA::leaveMulticastGroup(Ieee80211DataOrMgmtFrame *frame)
{
    Ieee80211LBMSRequest *mframe = (Ieee80211LBMSRequest *) frame;
    EV <<parentModule()->fullPath()<<"."<<className()<<" LeaveMulticastGroup "<<mframe->getElement().multicastAddress<<endl;

    /*for (MulticastMACAddressListI j = multicastGroupList.begin (); j != multicastGroupList.end (); j++)
    {
         if (j->address == mframe->getElement().multicastAddress)
         {
              multicastGroupList.erase(j);
              continue;
         }
         break;
    }*/
    logMultiCast();
}

bool LbmsMacSTA::isNewGroup(Ieee80211Frame *frame)
{
    Ieee80211DataOrMgmtFrame *mframe = (Ieee80211DataOrMgmtFrame *) frame;
    EV <<parentModule()->fullPath()<<"."<<className()<<"isNewGroup : Multicast address = "<<mframe->getAddress3()<<endl;
    if (multicastGroupList.empty()) return true;
    else 
    {
		for (MulticastMACAddressListI j = multicastGroupList.begin (); j != multicastGroupList.end (); j++)
		{
			if (j->address == mframe->getAddress3()) return false;
		}
    }
    return true;
}

void LbmsMacSTA::logMultiCast()
{
    EV << "--------------------------Log Multicast Member--------------------" <<endl;
    int i=0;
    if (multicastGroupList.empty()) return;
    for (MulticastMACAddressListI j = multicastGroupList.begin(); j != multicastGroupList.end(); j++,i++) {
         EV <<parentModule()->fullPath()<<"."<<className()<<":Groupe Address "<<i+1<<" "<<j->address<<endl;
         EV <<" isLeader " <<j->isLeader<<endl ;
    }
    EV << "-----------------------Log Multicast Member END--------------------" <<endl;
}

void LbmsMacSTA::processIncomingLBMS(Ieee80211Frame *frameToSend)
{
    EV<<"Handle LBMS Frame"<<endl;
    if (frameToSend->getType() == ST_LBMS_REPORT) setLeader(frameToSend);
    sendUp(frameToSend);
}

