//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2007-10-11
//
#include "LbmsMgmtSTA.h"
#include "Ieee802Ctrl_m.h"
#include "NotifierConsts.h"
#include "PhyControlInfo_m.h"
#include "RadioState.h"
#include "IGMPMessage_m.h"
#include "../../../Network/IPv4/IPDatagram_m.h"
#include "../../../Transport/UDP/UDPPacket.h"
#include "../../../Transport/RTP/RTPInnerPacket.h"

Define_Module(LbmsMgmtSTA);

// message kind values for timers
#define MK_AUTH_TIMEOUT         1
#define MK_ASSOC_TIMEOUT        2
#define MK_SCAN_SENDPROBE       3
#define MK_SCAN_MINCHANNELTIME  4
#define MK_SCAN_MAXCHANNELTIME  5
#define MK_BEACON_TIMEOUT       6
#define MK_LBMS_TIMER           7

#define MAX_BEACONS_MISSED 3.5  // beacon lost timeout, in beacon intervals (doesn't need to be integer)


std::ostream& operator<<(std::ostream& os, const LbmsMgmtSTA::ScanningInfo& scanning)
{
    os << "activeScan=" << scanning.activeScan
       << " probeDelay=" << scanning.probeDelay
       << " curChan=" << scanning.channelList[scanning.currentChannelIndex]
       << " minChanTime=" << scanning.minChannelTime
       << " maxChanTime=" << scanning.maxChannelTime;
    os << " chanList={";
    for (int i=0; i<scanning.channelList.size(); i++)
        os << (i==0 ? "" : " ") << scanning.channelList[i];
    os << "}";

    return os;
}

std::ostream& operator<<(std::ostream& os, const LbmsMgmtSTA::APInfo& ap)
{
    os << "AP addr=" << ap.address
       << " chan=" << ap.channel
       << " ssid=" << ap.ssid
       //TBD supportedRates
       << " beaconIntvl=" << ap.beaconInterval
       << " rxPower=" << ap.rxPower
       << " authSeqExpected=" << ap.authSeqExpected
       << " isAuthenticated=" << ap.isAuthenticated;
    return os;
}

std::ostream& operator<<(std::ostream& os, const LbmsMgmtSTA::AssociatedAPInfo& assocAP)
{
    os << "AP addr=" << assocAP.address
       << " chan=" << assocAP.channel
       << " ssid=" << assocAP.ssid
       << " beaconIntvl=" << assocAP.beaconInterval
       << " receiveSeq="  << assocAP.receiveSequence
       << " rxPower=" << assocAP.rxPower;
    return os;
}

void LbmsMgmtSTA::initialize(int stage)
{
    Ieee80211MgmtSTA::initialize(stage);
    if (stage==0)
    {
        lbmsInterval = par ("lbmsInterval");
    }
}

void LbmsMgmtSTA::handleTimer(cMessage *msg)
{
    if (msg->kind()==MK_AUTH_TIMEOUT)
    {
        // authentication timed out
        APInfo *ap = (APInfo *)msg->contextPointer();
        ev << "Authentication timed out, AP address = " << ap->address << "\n";

        // send back failure report to agent
        sendAuthenticationConfirm(ap, PRC_TIMEOUT);
    }
    else if (msg->kind()==MK_ASSOC_TIMEOUT)
    {
        // association timed out
        APInfo *ap = (APInfo *)msg->contextPointer();
        ev << "Association timed out, AP address = " << ap->address << "\n";

        // send back failure report to agent
        sendAssociationConfirm(ap, PRC_TIMEOUT);
    }
    else if (msg->kind()==MK_SCAN_MAXCHANNELTIME)
    {
        // go to next channel during scanning
        bool done = scanNextChannel();
        if (done)
            sendScanConfirm(); // send back response to agents' "scan" command
        delete msg;
    }
    else if (msg->kind()==MK_SCAN_SENDPROBE)
    {
        // Active Scan: send a probe request, then wait for minChannelTime (11.1.3.2.2)
        delete msg;
        sendProbeRequest();
        cMessage *timerMsg = new cMessage("minChannelTime", MK_SCAN_MINCHANNELTIME);
        scheduleAt(simTime()+scanning.minChannelTime, timerMsg); //XXX actually, we should start waiting after ProbeReq actually got transmitted
    }
    else if (msg->kind()==MK_SCAN_MINCHANNELTIME)
    {
        // Active Scan: after minChannelTime, possibly listen for the remaining time until maxChannelTime
        delete msg;
        if (scanning.busyChannelDetected)
        {
            ev << "Busy channel detected during minChannelTime, continuing listening until maxChannelTime elapses\n";
            cMessage *timerMsg = new cMessage("maxChannelTime", MK_SCAN_MAXCHANNELTIME);
            scheduleAt(simTime()+scanning.maxChannelTime - scanning.minChannelTime, timerMsg);
        }
        else
        {
            ev << "Channel was empty during minChannelTime, going to next channel\n";
            bool done = scanNextChannel();
            if (done)
                sendScanConfirm(); // send back response to agents' "scan" command
        }
    }
    else if (msg->kind()==MK_BEACON_TIMEOUT)
    {
        // missed a few consecutive beacons
        beaconLost();
    }
    else if (msg->kind()==MK_LBMS_TIMER)
    {
        sendLbmsReport();
        scheduleAt(simTime()+dblrand()*lbmsInterval, lbmsTimer);
    }
    else
    {
        error("internal error: unrecognized timer '%s'", msg->name());
    }
}

void LbmsMgmtSTA::handleUpperMessage(cMessage *msg)
{
    ev<< "handleUpperMessage "<<msg->name()<<endl;
    if (!strcmp(msg->name(),"ReceiverStats"))
    {
        UDPPacket *UDPpayload = check_and_cast<UDPPacket *>(msg->decapsulate());
        delete msg;
        RTPInnerPacket *payload = check_and_cast<RTPInnerPacket *>(UDPpayload->decapsulate());
        delete UDPpayload;
        RTPlost = payload->getLostPercentage();
        delete payload;
        return;
    }
    else if (!strcmp(msg->name(), "Join"))
    {
        UDPPacket *UDPpayload = check_and_cast<UDPPacket *>(msg->decapsulate());
        delete msg;
        IGMPMessage *payload = check_and_cast<IGMPMessage *>(UDPpayload->decapsulate());
        delete UDPpayload;
        lbmsTimer = new cMessage("lbmsTimer", MK_LBMS_TIMER);
        scheduleAt(simTime()+dblrand()*lbmsInterval, lbmsTimer);
        return sendLbmsRequest(payload);
    }
    else if (!strcmp(msg->name(), "Leave"))
    {
        UDPPacket *UDPpayload = check_and_cast<UDPPacket *>(msg->decapsulate());
        delete msg;
        IGMPMessage *payload = check_and_cast<IGMPMessage *>(UDPpayload->decapsulate());
        delete UDPpayload;
        cancelAndDelete(lbmsTimer);
        return sendLbmsRequest(payload);
    }

    Ieee80211DataFrame *frame = encapsulate(msg);
    sendOrEnqueue(frame);
}

void LbmsMgmtSTA::sendAuthenticationConfirm(APInfo *ap, int resultCode)
{
    Ieee80211Prim_AuthenticateConfirm *confirm = new Ieee80211Prim_AuthenticateConfirm();
    confirm->setAddress(ap->address);
    sendConfirm(confirm, resultCode);
}

void LbmsMgmtSTA::sendAssociationConfirm(APInfo *ap, int resultCode)
{
    sendConfirm(new Ieee80211Prim_AssociateConfirm(), resultCode);
}

void LbmsMgmtSTA::sendLbmsRequest(cMessage *msg)
{
    ev<<"Sending LBMS Request"<<endl;
    IGMPMessage *payload = check_and_cast<IGMPMessage *>(msg);
    Ieee80211LBMSRequest * lbmsRequest;
    if (payload->getType() ==  MEMBERSHIP_REPORT)
        lbmsRequest = new Ieee80211LBMSRequest("Join");
    else if (payload->getType() ==  LEAVE_GROUP)
        lbmsRequest = new Ieee80211LBMSRequest("Leave");

    unsigned char macBytes[6];
    macBytes[0] = 0x01;
    macBytes[1] = 0x00;
    macBytes[2] = 0x5e;
    macBytes[3] = payload->getGroupAddress().getDByte(1) & 0x7f;
    macBytes[4] = payload->getGroupAddress().getDByte(2);
    macBytes[5] = payload->getGroupAddress().getDByte(3);
    lbmsRequest->getElement().multicastAddress.setAddressBytes(macBytes);
    lbmsRequest->getElement().MulticastACKPolicy=0;
    lbmsRequest->getElement().RetryLimit=3;
    sendOrEnqueue(lbmsRequest);
}

void LbmsMgmtSTA::sendLbmsReport()
{
    ev<<"Sending LBMS Report"<<endl;
    Ieee80211LBMSReport * lbmsReport = new Ieee80211LBMSReport("LbmsReport");
    lbmsReport->setElementArraySize(1);
    lbmsReport->setPacketLoss(RTPlost);
    sendOrEnqueue(lbmsReport);
}

void LbmsMgmtSTA::handleLbmsRequestFrame(Ieee80211LBMSRequest *frame)
{
    ev<<"HandleLBMSRequestFrame"<<endl;
    delete frame;
}

void LbmsMgmtSTA::handleLbmsReportFrame(Ieee80211LBMSReport *frame)
{
    ev<<"HandleLBMSReportFrame"<<endl;
    delete frame;
}
