//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2008-01-11
//

#include <omnetpp.h>

#include "defs.h"

#include "IPAddress.h"
#include "UDPSocket.h"
#include "UDPControlInfo_m.h"

#include "types.h"
#include "RTCPCoordinatorModule.h"
#include "RTPInnerPacket.h"
#include "RTPParticipantInfo.h"
#include "RTPSenderInfo.h"
#include "RTPReceiverInfo.h"

Define_Module_Like(RTCPCoordinatorModule, RTCPModuleWithRateControl);


RTCPCoordinatorModule::~RTCPCoordinatorModule()
{
}

void RTCPCoordinatorModule::initialize()
{
	RTCPEndsystemModule::initialize();
    
	_packetsCalculated = 0;
    _averagePacketSize = 0.0;
    _rtpPacketsSent = 0;
    _rtpAveragePacketSize = 0.0;

	_lsr = 0.0;
	_ntpSender = 0.0;
	_rtt = 0.0;
	_fractionLost = 0.0;

	_stategie = 0 ;	// 0 : the best case
					// 1 : the worst case
					// 2 : The medium case
};

void RTCPCoordinatorModule::handleSelfMessage(cMessage *msg)
{
	// it's time to create an rtcp packet
	if (!_ssrcChosen) {
		chooseSSRC();
		RTPInnerPacket *rinp1 = new RTPInnerPacket("rtcpInitialized()");
		rinp1->rtcpInitialized(_senderInfo->ssrc());
		send(rinp1, "toRTP");
	}

	createPacket();

	if (!_leaveSession) {
		scheduleInterval();
	}
};

void RTCPCoordinatorModule::connectRet()
{
    // schedule first rtcp packet
    double intervalLength = 2.5 * (dblrand() + 0.5);
    cMessage *reminderMessage = new cMessage("Interval");
    scheduleAt(simTime() + intervalLength, reminderMessage);
};



void RTCPCoordinatorModule::createPacket()
{
	// first packet in an rtcp compound packet must
    // be a sender or receiver report
    RTCPReceiverReportPacket *reportPacket;

    // if this rtcp end system is a sender (see SenderInformation::isSender() for
    // details) insert a sender report
    if (_senderInfo && _senderInfo->isSender()) {
        RTCPSenderReportPacket *senderReportPacket = new RTCPSenderReportPacket("SenderReportPacket");
        senderReportPacket->setSenderReport(_senderInfo->senderReport(simTime()));
        reportPacket = senderReportPacket;
    }
    else
	{
        reportPacket = new RTCPReceiverReportPacket("ReceiverReportPacket");
		ReceptionReport *report = _receiverInfo->receptionReport(simTime());
		//report->writeContents();
		if (_stategie == 0)
			computeBestCondition();
		else if (_stategie == 1)
			computeMediumCondition();
		else computeWorstCondition();

		report->setDelaySinceLastSR(report->delaySinceLastSR() - _rtt);
		u_int8 fractionLost = (u_int8)(_fractionLost*256);
		report->setFractionLost(fractionLost);
		reportPacket->addReceptionReport(report);
	}
    
	if (_senderInfo&& _senderInfo->isSender())
		reportPacket->setSSRC(_senderInfo->ssrc());
	else 
		reportPacket->setSSRC(_receiverInfo->ssrc());

	// insert receiver reports for packets from other sources
    // insert source description items (at least common name)
    RTCPSDESPacket *sdesPacket = new RTCPSDESPacket("SDESPacket");
	SDESChunk *chunk;
	if (_senderInfo && _senderInfo->isSender()) chunk = _senderInfo->sdesChunk();
	else chunk = _receiverInfo->sdesChunk();
	sdesPacket->addSDESChunk(chunk);
	

    RTCPCompoundPacket *compoundPacket = new RTCPCompoundPacket("RTCPCompoundPacket");

    compoundPacket->addRTCPPacket(reportPacket);
    compoundPacket->addRTCPPacket(sdesPacket);

    // create rtcp app/bye packets if needed
    if (_leaveSession) {
        RTCPByePacket *byePacket = new RTCPByePacket("ByePacket");
        if (_senderInfo)
			byePacket->setSSRC(_senderInfo->ssrc());
		else 
			byePacket->setSSRC(_receiverInfo->ssrc());
        compoundPacket->addRTCPPacket(byePacket);
    };

	calculateAveragePacketSize(compoundPacket->length());

    cMessage *msg = new cMessage("RTCPCompoundPacket");
    msg->encapsulate(compoundPacket);
    msg->setKind(UDP_C_DATA);
    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSockId(_socketFdOut);
    ctrl->setDestAddr(_destinationAddress);
    ctrl->setDestPort(_port);
    msg->setControlInfo(ctrl);

    send(msg, "toUDP");

    if (_leaveSession) {
        RTPInnerPacket *rinp = new RTPInnerPacket("sessionLeft()");
        rinp->sessionLeft();
        send(rinp, "toRTP");
    };
};

void RTCPCoordinatorModule::processOutgoingRTPPacket(RTPPacket *packet)
{
    _rtpAveragePacketSize = ((double)(_rtpPacketsSent * _rtpAveragePacketSize + (double)packet->length()) / (double)(_rtpPacketsSent+1));
	_rtpPacketsSent++;
	RTCPEndsystemModule::processOutgoingRTPPacket(packet);
};


void RTCPCoordinatorModule::processIncomingRTCPPacket(RTCPCompoundPacket *packet, IN_Addr address, IN_Port port)
{	
	calculateAveragePacketSize(packet->length());
    cArray *rtcpPackets = packet->rtcpPackets();

    simtime_t arrivalTime = packet->arrivalTime();
    delete packet;
	
	for (int i = 0; i < rtcpPackets->items(); i++) {
        if (rtcpPackets->exist(i)) {
            // remove the rtcp packet from the rtcp compound packet
            RTCPPacket *rtcpPacket = (RTCPPacket *)(rtcpPackets->remove(i));

            if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_SR)
			{
				ev << "################SR###############";
                RTCPSenderReportPacket *rtcpSenderReportPacket = (RTCPSenderReportPacket *)rtcpPacket;
				_lsr = simTime();
				_ntpSender = ((double)rtcpSenderReportPacket->senderReport()->ntpTimeStamp()) / (65536.0 * 65536.0 );
				_receiverInfo->processSenderReport(rtcpSenderReportPacket->senderReport(), simTime());

            }
            else if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_RR)
			{
                RTCPReceiverReportPacket *rtcpReceiverReportPacket = (RTCPReceiverReportPacket *)rtcpPacket;
                u_int32 ssrc = rtcpReceiverReportPacket->ssrc();
				if (isNewMobileReceiver(ssrc))
				{
					addNewMobileReceiver(ssrc);
				}
				processMobileReceiverReport(rtcpReceiverReportPacket);
            }
            else if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_SDES)
			{
                RTCPSDESPacket *rtcpSDESPacket = (RTCPSDESPacket *)rtcpPacket;

                cArray *sdesChunks = rtcpSDESPacket->sdesChunks();

                for (int j = 0; j < sdesChunks->items(); j++)
				{
					if (sdesChunks->exist(j)) {
                        // remove the sdes chunk from the cArray of sdes chunks
                        SDESChunk *sdesChunk = (SDESChunk *)(sdesChunks->remove(j));
                        // this is needed to avoid seg faults
                        //sdesChunk->setOwner(this);
                        u_int32 ssrc = sdesChunk->ssrc();
						delete sdesChunk;
					}
                }

                delete sdesChunks;

            }
            else if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_BYE)
			{
                RTCPByePacket *rtcpByePacket = (RTCPByePacket *)rtcpPacket;
                u_int32 ssrc = rtcpByePacket->ssrc();
				removeMobileReceiver(ssrc);
				delete rtcpByePacket;

            }
            else {
                // app rtcp packets
            }
        delete rtcpPacket;
        }
    }
    delete rtcpPackets;
};


bool RTCPCoordinatorModule::isNewMobileReceiver(u_int32 ssrc)
{
    if (mobileReceiverInfoList.empty()) return true;
    else 
    {
		for (MobileReceiverInfoListI j = mobileReceiverInfoList.begin (); j != mobileReceiverInfoList.end (); j++)
		{
			if (j->ssrc == ssrc) return false;
		}
    }
    return true;
};


void RTCPCoordinatorModule::addNewMobileReceiver(u_int32 ssrc)
{
    MobileReceiverInfo mobileReceiverInfo;
    mobileReceiverInfo.ssrc = ssrc;
    mobileReceiverInfo.fractionloss = 0.0;
	mobileReceiverInfo.rtt = 0.0;
	mobileReceiverInfo.lsr = 0.0;
    mobileReceiverInfoList.push_back(mobileReceiverInfo);

    myMobileReceiverInfo();
};


void RTCPCoordinatorModule::removeMobileReceiver(u_int32 ssrc)
{
    for (MobileReceiverInfoListI j = mobileReceiverInfoList.begin(); j != mobileReceiverInfoList.end(); j++)
    {
		if(j->ssrc == ssrc)
			mobileReceiverInfoList.erase(j);
    }

    myMobileReceiverInfo();
};


void RTCPCoordinatorModule::myMobileReceiverInfo()
{
	ev<< "------------------------Log Multicast Group ----------------------" <<endl;
    int l=1;
    for (MobileReceiverInfoListI j = mobileReceiverInfoList.begin (); j != mobileReceiverInfoList.end (); j++)
	{
         ev<<"Membre "<< l++ <<" ssrc:" << j->ssrc;
         ev<<" Packet loss "<<j->fractionloss;
         ev<<" Rtt "<<j->rtt;
		 ev<<" Lsr "<<j->lsr;
         ev<< endl;
    }
};

void RTCPCoordinatorModule::processMobileReceiverReport(RTCPReceiverReportPacket *rtcpReceiverReportPacket)
{
	u_int32 ssrc = rtcpReceiverReportPacket->ssrc();

	cArray *receptionReports = rtcpReceiverReportPacket->receptionReports();
	ev << " receptionReportsItems "  << receptionReports->items() << endl;
    for (int j = 0; j < receptionReports->items(); j++) 
	{
        ReceptionReport *receptionReport = (ReceptionReport *)(receptionReports->remove(j));
		double rtt = calculateRoundTripTime(receptionReport);
		double packetloss = calculatePacketLoss(receptionReport);

	    for (MobileReceiverInfoListI j = mobileReceiverInfoList.begin(); j != mobileReceiverInfoList.end(); j++)
		{
			if(j->ssrc == ssrc)
			{
				if (rtt >0) j->rtt = rtt;
				j->fractionloss = packetloss;
				j->lsr = ((double)receptionReport->lastSR())/65536.0;
			}
		}

		delete  receptionReport ;
    }
	myMobileReceiverInfo();
};

double RTCPCoordinatorModule::calculateRoundTripTime(ReceptionReport *receptionReport)
{
	double rtt;
	double dlsr = ((double)receptionReport->delaySinceLastSR())/65536.0;
	double lsr = ((double)receptionReport->lastSR())/65536.0;
	ev  << "simTime() " << simTime() << " dlsr " << dlsr << " _ntpSender " << _ntpSender << " lsr " << lsr << endl;
	if ((abs(lsr - _ntpSender))<0.1)
		rtt = simTime() - dlsr - _lsr;
	else rtt = 0.0;
	ev << " rtt " << rtt << endl;
	return rtt;
};

double RTCPCoordinatorModule::calculatePacketLoss(ReceptionReport *receptionReport)
{
	double p = ((double)receptionReport->fractionLost())/256;
	return p;
};

void RTCPCoordinatorModule::computeWorstCondition()
{
	double rtt = 0.0;
	double factionloss = 0.0;
    for (MobileReceiverInfoListI j = mobileReceiverInfoList.begin (); j != mobileReceiverInfoList.end (); j++)
	{
        if (j->rtt > rtt);
		rtt = j->rtt;
		if (j->fractionloss > factionloss)
		factionloss = j->fractionloss;
    }

	_rtt = rtt;
	_fractionLost = factionloss;
	ev << " New RTT "  << _rtt << " New FractionLost " << _fractionLost << endl;
}


void RTCPCoordinatorModule::computeMediumCondition()
{
	double rtt = 0.0;
	double factionloss = 0.0;
	int i = 0;
    for (MobileReceiverInfoListI j = mobileReceiverInfoList.begin (); j != mobileReceiverInfoList.end (); j++)
	{
		i++;
        rtt = ((rtt * (i-1)) + j->rtt )/i;
		factionloss = ((factionloss * (i-1))+j->fractionloss)/i;
    }

	_rtt = rtt;
	_fractionLost = factionloss;
	ev << " New RTT "  << _rtt << " New FractionLost " << _fractionLost << endl;
}

void RTCPCoordinatorModule::computeBestCondition()
{
	double rtt = mobileReceiverInfoList.begin()->rtt;
	double factionloss = 1;
    for (MobileReceiverInfoListI j = mobileReceiverInfoList.begin(); j != mobileReceiverInfoList.end (); j++)
	{
		std :: cout << " Ssrc = " << j->ssrc << " rtt = " << j->rtt << " fractionloss " << j->fractionloss << endl;		
		if (j->fractionloss < factionloss)
		{
			factionloss = j->fractionloss;
			rtt = j->rtt;
		}
    }

	_rtt = rtt;
	_fractionLost = factionloss;
	std::cout << "computeBestCondition: New RTT "  << _rtt << " New FractionLost " << _fractionLost << endl;
};