//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2008-01-11
//

#include <omnetpp.h>

#include "defs.h"

#include "IPAddress.h"
#include "UDPSocket.h"
#include "UDPControlInfo_m.h"

#include "types.h"
#include "RTCPwithRateControl.h"
#include "RTPInnerPacket.h"
#include "RTPParticipantInfo.h"
#include "RTPSenderInfo.h"
#include "RTPReceiverInfo.h"

Define_Module_Like(RTCPwithRateControl, RTCPModuleWithRateControl);


RTCPwithRateControl::~RTCPwithRateControl()
{
}

void RTCPwithRateControl::initialize()
{
	RTCPEndsystemModule::initialize();
    
	_packetsCalculated = 0;
    _averagePacketSize = 0.0;
    _rtpPacketsSent = 0;
    _rtpAveragePacketSize = 0.0;
	_rTcp = 0;
	_bandwidth = 64000;
	_packetloss = 0;
	_rtt = 0;

	_rateAdaptationAlgorithm = par("rateAdaptationAlgorithm");
	if (_rateAdaptationAlgorithm > 0)
	{
		_air = _airInitial = par("air");
		_reductionfactor = par("reductionFactor");
	}
	_rttOutVector.setName("RTT");
	_fractionLostOutVector.setName("Fraction Lost");
	_recvRateOutVector.setName("Received Rate");
	_packetSizeOutVector.setName("Packet Size");
	_minRate = par("minRate");
	_maxRate = par("maxRate");
};

//
// methods for different messages
//

void RTCPwithRateControl::initializeRTCP(RTPInnerPacket *rinp)
{
    _mtu = rinp->mtu();
    _bandwidth = rinp->bandwidth();
    _rtcpPercentage = rinp->rtcpPercentage();
    _destinationAddress = rinp->address();
    _port = rinp->port();

    _senderInfo = new RTPSenderInfo();
	SDESItem *sdesItem1 = new SDESItem(SDESItem::SDES_CNAME, rinp->commonName());
    _senderInfo->addSDESItem(sdesItem1);

    // create server socket for receiving rtcp packets
    createSocket();
};


void RTCPwithRateControl::handleSelfMessage(cMessage *msg)
{
	// it's time to create an rtcp packet
	if (!_ssrcChosen) {
		chooseSSRC();
		RTPInnerPacket *rinp1 = new RTPInnerPacket("rtcpInitialized()");
		rinp1->rtcpInitialized(_senderInfo->ssrc());
		send(rinp1, "toRTP");
	}

	createPacket();

	if (!_leaveSession) {
		scheduleInterval();
	}
};

void RTCPwithRateControl::connectRet()
{
    // schedule first rtcp packet
    double intervalLength = 2.5 * (dblrand() + 0.5);
    cMessage *reminderMessage = new cMessage("Interval");
    scheduleAt(simTime() + intervalLength, reminderMessage);
};



void RTCPwithRateControl::createPacket()
{
	// first packet in an rtcp compound packet must
    // be a sender or receiver report

    RTCPSenderReportPacket *reportPacket = new RTCPSenderReportPacket("SenderReportPacket");
    reportPacket->setSenderReport(_senderInfo->senderReport(simTime()));

	reportPacket->setSSRC(_senderInfo->ssrc());

	// insert receiver reports for packets from other sources
    // insert source description items (at least common name)
    RTCPSDESPacket *sdesPacket = new RTCPSDESPacket("SDESPacket");
	SDESChunk *chunk;
	chunk = _senderInfo->sdesChunk();
	sdesPacket->addSDESChunk(chunk);
	

    RTCPCompoundPacket *compoundPacket = new RTCPCompoundPacket("RTCPCompoundPacket");

    compoundPacket->addRTCPPacket(reportPacket);
    compoundPacket->addRTCPPacket(sdesPacket);

    // create rtcp app/bye packets if needed
    if (_leaveSession) {
        RTCPByePacket *byePacket = new RTCPByePacket("ByePacket");
		byePacket->setSSRC(_senderInfo->ssrc());
        compoundPacket->addRTCPPacket(byePacket);
    };

	calculateAveragePacketSize(compoundPacket->length());

    cMessage *msg = new cMessage("RTCPCompoundPacket");
    msg->encapsulate(compoundPacket);
    msg->setKind(UDP_C_DATA);
    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSockId(_socketFdOut);
    ctrl->setDestAddr(_destinationAddress);
    ctrl->setDestPort(_port);
    msg->setControlInfo(ctrl);

    send(msg, "toUDP");

    if (_leaveSession) {
        RTPInnerPacket *rinp = new RTPInnerPacket("sessionLeft()");
        rinp->sessionLeft();
        send(rinp, "toRTP");
    };
};

void RTCPwithRateControl::processOutgoingRTPPacket(RTPPacket *packet)
{
    _rtpAveragePacketSize = ((double)(_rtpPacketsSent * _rtpAveragePacketSize + (double)packet->length()) / (double)(_rtpPacketsSent+1));
	_rtpPacketsSent++;
	RTCPEndsystemModule::processOutgoingRTPPacket(packet);
};


void RTCPwithRateControl::processIncomingRTCPPacket(RTCPCompoundPacket *packet, IN_Addr address, IN_Port port)
{	
	std::cout << " RTCPwithRateControl::processIncomingRTCPPacket "<< endl;
	calculateAveragePacketSize(packet->length());
    cArray *rtcpPackets = packet->rtcpPackets();

    simtime_t arrivalTime = packet->arrivalTime();
    delete packet;
	
	for (int i = 0; i < rtcpPackets->items(); i++) {
        if (rtcpPackets->exist(i)) {
            // remove the rtcp packet from the rtcp compound packet
            RTCPPacket *rtcpPacket = (RTCPPacket *)(rtcpPackets->remove(i));

            if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_SR)
			{
                RTCPSenderReportPacket *rtcpSenderReportPacket = (RTCPSenderReportPacket *)rtcpPacket;
				_senderInfo->processSenderReport(rtcpSenderReportPacket->senderReport(), simTime());

            }
            else if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_RR)
			{
                RTCPReceiverReportPacket *rtcpReceiverReportPacket = (RTCPReceiverReportPacket *)rtcpPacket;
                u_int32 ssrc = rtcpReceiverReportPacket->ssrc();
                RTPReceiverInfo *participantInfo = (RTPReceiverInfo *)findParticipantInfo(ssrc);
                if (participantInfo == NULL) {
                    participantInfo = new RTPReceiverInfo(ssrc);
                    participantInfo->setAddress(address);
                    participantInfo->setRTCPPort(port);
					participantInfo->setFreidly(0);
                    _participantInfos->add(participantInfo);
                }
                else {
                    if (participantInfo->address() == address) {
                        if (participantInfo->rtcpPort() == IPSuite_PORT_UNDEF) {
                            participantInfo->setRTCPPort(port);
                        }
                        else {
                            // check for ssrc conflict
                        }
                    }
                    else {
                        // check for ssrc conflict
                    };
                }

                cArray *receptionReports = rtcpReceiverReportPacket->receptionReports();
                for (int j = 0; j < receptionReports->items(); j++) {
                    if (receptionReports->exist(j)) {
                        ReceptionReport *receptionReport = (ReceptionReport *)(receptionReports->remove(j));
						if (receptionReport->lastSR())
							updateTransmissionRate(receptionReport);
						participantInfo->processReceptionReport(receptionReport, simTime());
						_lastReceivedReport = arrivalTime;
                    }
                };
                delete receptionReports;
            }
            else if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_SDES)
			{
                RTCPSDESPacket *rtcpSDESPacket = (RTCPSDESPacket *)rtcpPacket;

                cArray *sdesChunks = rtcpSDESPacket->sdesChunks();

                for (int j = 0; j < sdesChunks->items(); j++) {
                    if (sdesChunks->exist(j)) {
                        // remove the sdes chunk from the cArray of sdes chunks
                        SDESChunk *sdesChunk = (SDESChunk *)(sdesChunks->remove(j));
                        // this is needed to avoid seg faults
                        u_int32 ssrc = sdesChunk->ssrc();
                        RTPParticipantInfo *participantInfo = findParticipantInfo(ssrc);
                        if (participantInfo == NULL) {
                            participantInfo = new RTPReceiverInfo(ssrc);
                            participantInfo->setAddress(address);
                            participantInfo->setRTCPPort(port);
                            _participantInfos->add(participantInfo);
                        }
                        else {
                            // check for ssrc conflict
                        }
                        participantInfo->processSDESChunk(sdesChunk, arrivalTime);
                    }
                }
                delete sdesChunks;

            }
            else if (rtcpPacket->packetType() == RTCPPacket::RTCP_PT_BYE)
			{
                RTCPByePacket *rtcpByePacket = (RTCPByePacket *)rtcpPacket;
                u_int32 ssrc = rtcpByePacket->ssrc();
                RTPParticipantInfo *participantInfo = findParticipantInfo(ssrc);

                if (participantInfo != NULL && participantInfo != _senderInfo && participantInfo != _senderInfo) {
                    _participantInfos->remove(participantInfo);
                    delete participantInfo;
                    // perhaps it would be useful to inform
                    // the profile to remove the corresponding
                    // receiver module
                };
            }
            else {
                // app rtcp packets
            }
        delete rtcpPacket;
        }
    }
    delete rtcpPackets;
};

void RTCPwithRateControl::calculateAveragePacketSize(int size) {
    // add size of ip and udp header to given size before calculating
    _averagePacketSize = ((double)(_packetsCalculated) * _averagePacketSize + (double)(size + 20 + 8)) / (double)(_packetsCalculated+1);
	_packetsCalculated++;	
};


void RTCPwithRateControl::updateTransmissionRate(ReceptionReport *receptionReport)
{
	std::cout << " RTCPwithRateControl::updateTransmissionRate ";
	if ( _rtt == 0)
		_rtt = calculateRoundTripTime(receptionReport);
	else 
		_rtt = 0.5 *_rtt + 0.5 * calculateRoundTripTime(receptionReport);

	_rtt = calculateRoundTripTime(receptionReport);
	//_rttOutVector.record(_rtt);

	_packetloss = calculatePacketLoss(receptionReport);
	//_fractionLostOutVector.record(_packetloss);

	double rate = calculateRate(_packetloss, _rtt);

	_bandwidth = min ( max (rate , _minRate) , _maxRate);

	if (_senderInfo && _senderInfo->isSender() && (!_leaveSession) && (_rateAdaptationAlgorithm > 0))
	{
		RTPInnerPacket *rinp = new RTPInnerPacket("ChangeSessionBandwidth");
		rinp->changeBandwidth(_bandwidth);
		send(rinp, "toRTP");
	};
}

double RTCPwithRateControl::calculateRate(double p, double rtt)
{
	if (_rateAdaptationAlgorithm == 0)
		return _bandwidth;
	if (_rateAdaptationAlgorithm == 1)
		return rateAdaptationAlgorithm(p);
	else 
	if (_rateAdaptationAlgorithm == 2)
		return directAdjustementAlgorithm(p,rtt);
	else
	if (_rateAdaptationAlgorithm == 3)
		return lostDelayBasedAdjustementAlgorithm(p,rtt);
	else
	if (_rateAdaptationAlgorithm == 4)
		return lostDelayBasedAdjustementAlgorithmPlus(p,rtt);
	else
	if (_rateAdaptationAlgorithm == 5)
		return TRFC(p,rtt);
	else error(" Error in specifying Rate Adaptation Algorithm");
}

double RTCPwithRateControl::rateAdaptationAlgorithm(double p)
{
	double r;
	if (p > 0) // loss situation
		r = _bandwidth / 2;
	else // no loss situation
		r = _bandwidth * 2;
	return r;
}

double RTCPwithRateControl::directAdjustementAlgorithm(double p, double rtt)
{
	double r;
	
	if (p == 0) // no loss situation
	{
		_rTcp = calculateThroughtput( _rtpAveragePacketSize , rtt, p);
		double rAdd = _bandwidth + _air;
		r = min( _rTcp , rAdd);
	}
	else // loss situation
		r = _bandwidth/2;
	return r;
}

double RTCPwithRateControl::lostDelayBasedAdjustementAlgorithm(double p, double rtt)
{
	double r = 0;
	double bottleneckBandwidth = 1000000; // should be mesured with Packet Pair method
	if (p > 0) /// loss situation
	{
		_rTcp = calculateThroughtput( _rtpAveragePacketSize , rtt,p);
		_air = _airInitial;
		r = _bandwidth * (1 - (p * _reductionfactor));
	}
	if (p == 0) /// no loss situation
	{
		_air *= (1 - ( _bandwidth /bottleneckBandwidth));
		r = _bandwidth + _air;
	}
	return r;
}

double RTCPwithRateControl::lostDelayBasedAdjustementAlgorithmPlus(double p, double rtt)
{
	double r;
	double bottleneckBandwidth = 1000000; // may be calculated with Packet Pair method
	if (p == 0) /// no loss situation
	{
		double aAdd = (2 - (double)(_bandwidth / bottleneckBandwidth)) * _air;
		double aExp = (1 - exp(-(1-(double)(_bandwidth / bottleneckBandwidth)))) * _bandwidth;
		double aTcp;
		if (_lastReceivedReport != 0)
			aTcp = _rtpAveragePacketSize * ((((simTime()-_lastReceivedReport) / rtt) + 1) / (2*rtt));
		else 
			aTcp = _rtpAveragePacketSize * (((5.0 / rtt) + 1) / (2*rtt));
		//std::cout<<" air "<<_air<<" Bandwidth "<< _bandwidth<<" aAdd "
		//	<<aAdd <<" aExp "<< aExp <<" aTcp "<<aTcp<<endl; 
		double temp = aAdd;
		if (aExp < aAdd )
			temp = aExp; 
		if (aTcp < temp)
			temp = aTcp;
		_air = temp;
		r = _bandwidth + _air;
	}
	else /// loss situation
	{
		_rTcp = calculateThroughtput((double)_rtpAveragePacketSize, rtt, p);
		r = (double)_bandwidth * (1-sqrt(p));
		if (_rTcp > r)
			r = _rTcp;
		_air = _airInitial;
	}
	//std::cout<<"New rate = " << r<<endl;
	return r;
}

double RTCPwithRateControl::TRFC(double p, double rtt) /// [RFC 3448]
{
	double r = 0; // new rate
	double temp = 0;
	if (p > 0) /// loss situation
	{
		_rTcp = calculateThroughtput( _rtpAveragePacketSize , rtt, p);
		assert( _rTcp > 0);
		temp = _rTcp;
		r = temp;
		if(temp < ((double) _rtpAveragePacketSize/64))
			r = ((double) _rtpAveragePacketSize/64);
	}
	if (p == 0) /// no loss situation
	{
		if ( temp < (_rtpAveragePacketSize/rtt))
			temp = (double)_rtpAveragePacketSize/rtt;
		r = temp;
	}
	return r;
}

double RTCPwithRateControl::calculateRoundTripTime(ReceptionReport *receptionReport)
{
	double rtt;
	double dlsr = ((double)receptionReport->delaySinceLastSR())/65536.0;
	double lsr  = ((double)receptionReport->lastSR())/65536.0;
	rtt = simTime() - dlsr - lsr;
	return rtt;
}

double RTCPwithRateControl::calculatePacketLoss(ReceptionReport *receptionReport)
{
	double p = ((double)receptionReport->fractionLost())/256.0;
	return p;
}


double RTCPwithRateControl::calculateThroughtput(int s, double r, double p)
{
	ev << simTime() << " size " << s << " rtt " << r << " loss " << p << endl;
	double temp = r*sqrt(2*p/3) + (4*r*(3*sqrt(3*p/8))*p*(1+32*p*p));
	ev << " TCP frindly" << (double) s / temp << endl;
	return  (((double) s) / temp);
}

void RTCPwithRateControl::chooseSSRC() {

    u_int32 ssrc = 0;
    bool ssrcConflict = false;
    do {
        ssrc = intrand(0x7fffffff);
        ssrcConflict = findParticipantInfo(ssrc) != NULL;
    } while (ssrcConflict);
	_senderInfo->setSSRC(ssrc);
	ev <<parentModule()->fullPath()<<"."<<className()<<":"<< " _senderInfo->setSSRC " << ssrc <<endl;
	_ssrcChosen = true;
};


void RTCPwithRateControl::processIncomingRTPPacket(RTPPacket *packet, IN_Addr address, IN_Port port)
{
	_senderInfo->processRTPPacket(packet, id(), simTime());
};