//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2008-01-11
//

#ifndef __RTCP_RATE_CONTROL_H__
#define __RTCP_RATE_CONTROL_H__

#include <omnetpp.h>

#include "defs.h"

#include "types.h"
#include "RTPInnerPacket.h"
#include "RTPParticipantInfo.h"
#include "RTPSenderInfo.h"
#include "RTPReceiverInfo.h"
#include "RTCPEndsystemModule.h"

/**
 * The class RTCPwithRateControl is responsible for creating, receiving and
 * processing of rtcp packets. It also keeps track of this and other
 * rtp end systems.
 */

class INET_API RTCPwithRateControl : public RTCPEndsystemModule
{
  protected:
        /**
         * Initializes variables.
         */
        virtual void initialize();

		virtual ~RTCPwithRateControl();
        /**
         * Handles self messages.
         */
        virtual void handleSelfMessage(cMessage *msg);
		virtual void initializeRTCP(RTPInnerPacket *rinp);

        /**
         * Called when the socket layer has finished a connect.
         */
        virtual void connectRet();

		/**
         * Tell the RTPEndSystem Module to adapt the transmission rate
         */
        virtual void updateTransmissionRate(ReceptionReport *receptionReport);

		virtual void chooseSSRC();

		virtual void processIncomingRTPPacket(RTPPacket *packet, IN_Addr address, IN_Port port);

    private:

        /**
         * The bandwidth for this rtp session.
         */
        int _bandwidth;

		/**
         * minimum send rate for this rtp session.
         */
        double _minRate;

		/**
         * maximum send rate for this rtp session.
         */
        double _maxRate;

		/**
         * The bandwidth for this rtp session.
         */
		int _rateAdaptationAlgorithm;

		/**
		 * The additive increase rate
		 */
		double _air;

		double _airInitial;

		double _rtt;
		double _packetloss;
		double _rTcp;
		double _lastReceivedReport;

		/**
		 * Tha reduction factor
		 */
		int _reductionfactor;

        /**
         * The percentage of bandwidth for rtcp.
         */
        int _rtcpPercentage;

        /**
         * The number of packets this rtcp module has
         * calculated.
         */
        int _packetsCalculated;

        /**
         * The average size of an RTCPCompoundPacket.
         */
        double _averagePacketSize;

		/**
         * The number of rtp packets sent
         */
	    int _rtpPacketsSent;

		/**
         * The average size of RTP packets sent by an RTP module.
         */
		double _rtpAveragePacketSize;

		/**
         * The output vector for sender TCP-freindly 
         */
        cOutVector _rttOutVector;
		
		/**
         * The output vector for sender TCP-freindly 
         */
        cOutVector _fractionLostOutVector;
		
		cOutVector _recvRateOutVector;
		
		cOutVector _packetSizeOutVector;

        /**
         * Creates and sends an RTCPCompoundPacket.
         */
        virtual void createPacket();

        /**
         * Extracts information of a sent RTPPacket.
         */
        virtual void processOutgoingRTPPacket(RTPPacket *packet);

        /**
         * Extracts information of a received RTCPCompoundPacket.
         */
        virtual void processIncomingRTCPPacket(RTCPCompoundPacket *packet, IN_Addr address, IN_Port port);

        /**
         * Recalculates the average size of an RTCPCompoundPacket when
         * one of this size has been sent or received.
         */
        virtual void calculateAveragePacketSize(int size);
		
		/**
         * Compute the transmission rate
         */
		double calculateRate(double p, double rtt);

		/**
         * Adjust the rate with LDA+ algorithm
         */
		double rateAdaptationAlgorithm(double p);

		/**
         * Adjust the rate with LDA+ algorithm
         */
		double lostDelayBasedAdjustementAlgorithmPlus(double p, double rtt);

		/**
         * Adjust the rate with LDA algorithm
         */
		double lostDelayBasedAdjustementAlgorithm(double p, double rtt);

		/**
         * Adjust the rate with DAA algorithm
         */
		double directAdjustementAlgorithm(double p, double rtt);

		double TRFC(double p, double rtt);

		double calculateRoundTripTime(ReceptionReport *receptionReport);

		double calculatePacketLoss(ReceptionReport *receptionReport);

		double calculateThroughtput(int s /*size*/, double r/*rtt*/, double p/*loss*/);
};

#endif


